"""Implementation of SQL Datastore."""
from typing import Callable, Dict, Iterable, List, Optional
from absl import logging

import sqlalchemy as sqla

from vizier.service import datastore
from vizier.service import key_value_pb2
from vizier.service import resources
from vizier.service import study_pb2
from vizier.service import vizier_oss_pb2
from google.longrunning import operations_pb2


# TODO: Consider using ORM API (when fixed) to reduce code length.
class SQLDataStore(datastore.DataStore):
  """SQL Datastore."""

  def __init__(self, engine):
    self._engine = engine
    self._connection = self._engine.connect()
    self._root_metadata = sqla.MetaData()
    self._owners_table = sqla.Table(
        'owners',
        self._root_metadata,
        sqla.Column('owner_name', sqla.String, primary_key=True),
    )
    self._studies_table = sqla.Table(
        'studies',
        self._root_metadata,
        sqla.Column('study_name', sqla.String, primary_key=True),
        sqla.Column('owner_id', sqla.String),
        sqla.Column('study_id', sqla.String),
        sqla.Column('serialized_study', sqla.String),
    )
    self._trials_table = sqla.Table(
        'trials',
        self._root_metadata,
        sqla.Column('trial_name', sqla.String, primary_key=True),
        sqla.Column('owner_id', sqla.String),
        sqla.Column('study_id', sqla.String),
        sqla.Column('trial_id', sqla.INTEGER),
        sqla.Column('serialized_trial', sqla.String),
    )
    self._suggestion_operations_table = sqla.Table(
        'suggestion_operations',
        self._root_metadata,
        sqla.Column('operation_name', sqla.String, primary_key=True),
        sqla.Column('owner_id', sqla.String),
        sqla.Column('client_id', sqla.String),
        sqla.Column('operation_number', sqla.INTEGER),
        sqla.Column('serialized_op', sqla.String),
    )
    self._early_stopping_operations_table = sqla.Table(
        'early_stopping_operations',
        self._root_metadata,
        sqla.Column('operation_name', sqla.String, primary_key=True),
        sqla.Column('owner_id', sqla.String),
        sqla.Column('study_id', sqla.String),
        sqla.Column('trial_id', sqla.INTEGER),
        sqla.Column('serialized_op', sqla.String),
    )
    self._root_metadata.create_all(self._engine)

  def create_study(self, study: study_pb2.Study) -> resources.StudyResource:
    study_resource = resources.StudyResource.from_name(study.name)
    owner_name = study_resource.owner_resource.name
    owner_query = self._owners_table.insert().values(owner_name=owner_name)
    try:
      self._connection.execute(owner_query)
    except sqla.exc.IntegrityError:
      logging.info('Owner with name %s currently exists.', owner_name)

    query = self._studies_table.insert().values(
        study_name=study.name,
        owner_id=study_resource.owner_id,
        study_id=study_resource.study_id,
        serialized_study=study.SerializeToString())
    try:
      self._connection.execute(query)
      return study_resource
    except sqla.exc.IntegrityError as integrity_error:
      raise datastore.AlreadyExistsError('Study with name %s already exists.' %
                                         study.name) from integrity_error

  def load_study(self, study_name: str) -> study_pb2.Study:
    query = sqla.select([self._studies_table])
    query = query.where(self._studies_table.c.study_name == study_name)
    result = self._connection.execute(query)
    row = result.fetchone()
    if not row:
      raise datastore.NotFoundError('Failed to find study name: %s' %
                                    study_name)

    return study_pb2.Study.FromString(row['serialized_study'])

  def delete_study(self, study_name: str) -> None:
    study_resource = resources.StudyResource.from_name(study_name)

    exists_query = sqla.select([self._studies_table])
    exists_query = exists_query.where(
        self._studies_table.c.study_name == study_name)
    exists_query = sqla.exists(exists_query).select()
    exists = self._connection.execute(exists_query).fetchone()[0]
    if not exists:
      raise datastore.NotFoundError('Study %s does not exist.' % study_name)

    query = self._studies_table.delete().where(
        self._studies_table.c.study_name == study_name)
    self._connection.execute(query)

    trial_query = self._trials_table.delete().where(
        self._trials_table.c.owner_id == study_resource.owner_id).where(
            self._trials_table.c.study_id == study_resource.study_id)
    self._connection.execute(trial_query)

  def list_studies(self, owner_name: str) -> List[study_pb2.Study]:
    owner_id = resources.OwnerResource.from_name(owner_name).owner_id

    exists_query = sqla.select([self._owners_table])
    exists_query = exists_query.where(
        self._owners_table.c.owner_name == owner_name)
    exists_query = sqla.exists(exists_query).select()
    exists = self._connection.execute(exists_query).fetchone()[0]
    if not exists:
      raise datastore.NotFoundError('Owner name %s does not exist.' %
                                    owner_name)

    query = sqla.select([self._studies_table])
    query = query.where(self._studies_table.c.owner_id == owner_id)
    result = self._connection.execute(query).fetchall()
    return [
        study_pb2.Study.FromString(row['serialized_study']) for row in result
    ]

  def create_trial(self, trial: study_pb2.Trial) -> resources.TrialResource:
    trial_resource = resources.TrialResource.from_name(trial.name)
    query = self._trials_table.insert().values(
        trial_name=trial.name,
        owner_id=trial_resource.owner_id,
        study_id=trial_resource.study_id,
        trial_id=trial_resource.trial_id,
        serialized_trial=trial.SerializeToString())
    try:
      self._connection.execute(query)
      return trial_resource
    except sqla.exc.IntegrityError as integrity_error:
      raise datastore.AlreadyExistsError('Trial with name %s already exists.' %
                                         trial.name) from integrity_error

  def get_trial(self, trial_name: str) -> study_pb2.Trial:
    query = sqla.select([self._trials_table])
    query = query.where(self._trials_table.c.trial_name == trial_name)
    result = self._connection.execute(query)
    row = result.fetchone()

    if not row:
      raise datastore.NotFoundError('Failed to find trial name: %s' %
                                    trial_name)

    return study_pb2.Trial.FromString(row['serialized_trial'])

  def update_trial(self, trial: study_pb2.Trial) -> resources.TrialResource:
    trial_resource = resources.TrialResource.from_name(trial.name)

    exists_query = sqla.select([self._trials_table])
    exists_query = exists_query.where(
        self._trials_table.c.trial_name == trial.name)
    exists_query = sqla.exists(exists_query).select()
    exists = self._connection.execute(exists_query).fetchone()[0]
    if not exists:
      raise datastore.NotFoundError('Trial %s does not exist.' % trial.name)

    query = sqla.update(self._trials_table).where(
        self._trials_table.c.trial_name == trial.name).values(
            trial_name=trial.name,
            owner_id=trial_resource.owner_id,
            study_id=trial_resource.study_id,
            trial_id=trial_resource.trial_id,
            serialized_trial=trial.SerializeToString())
    self._connection.execute(query)
    return trial_resource

  def list_trials(self, study_name: str) -> List[study_pb2.Trial]:
    study_resource = resources.StudyResource.from_name(study_name)

    exists_query = sqla.select([self._studies_table])
    exists_query = exists_query.where(
        self._studies_table.c.study_name == study_name)
    exists_query = sqla.exists(exists_query).select()
    exists = self._connection.execute(exists_query).fetchone()[0]
    if not exists:
      raise datastore.NotFoundError('Study name %s does not exist.' %
                                    study_name)

    query = sqla.select([self._trials_table])
    query = query.where(
        self._trials_table.c.owner_id == study_resource.owner_id)
    query = query.where(
        self._trials_table.c.study_id == study_resource.study_id)
    result = self._connection.execute(query)
    return [
        study_pb2.Trial.FromString(row['serialized_trial']) for row in result
    ]

  def delete_trial(self, trial_name: str) -> None:
    exists_query = sqla.select([self._trials_table])
    exists_query = exists_query.where(
        self._trials_table.c.trial_name == trial_name)
    exists_query = sqla.exists(exists_query).select()
    exists = self._connection.execute(exists_query).fetchone()[0]
    if not exists:
      raise datastore.NotFoundError('Trial %s does not exist.' % trial_name)

    query = self._trials_table.delete().where(
        self._trials_table.c.trial_name == trial_name)
    self._connection.execute(query)

  def max_trial_id(self, study_name: str) -> int:
    study_resource = resources.StudyResource.from_name(study_name)

    exists_query = sqla.select([self._studies_table])
    exists_query = exists_query.where(
        self._studies_table.c.study_name == study_name)
    exists_query = sqla.exists(exists_query).select()
    exists = self._connection.execute(exists_query).fetchone()[0]
    if not exists:
      raise datastore.NotFoundError('Study %s does not exist.' % study_name)

    query = sqla.select(
        [sqla.func.max(self._trials_table.c.trial_id, type_=sqla.INT)])
    query = query.where(
        self._trials_table.c.owner_id == study_resource.owner_id)
    query = query.where(
        self._trials_table.c.study_id == study_resource.study_id)
    potential_trial_id = self._connection.execute(query).fetchone()[0]
    if potential_trial_id is None:
      return 0
    return potential_trial_id

  def create_suggestion_operation(
      self, operation: operations_pb2.Operation
  ) -> resources.SuggestionOperationResource:
    resource = resources.SuggestionOperationResource.from_name(operation.name)
    query = self._suggestion_operations_table.insert().values(
        operation_name=operation.name,
        owner_id=resource.owner_id,
        client_id=resource.client_id,
        operation_number=resource.operation_number,
        serialized_op=operation.SerializeToString())

    try:
      self._connection.execute(query)
      return resource
    except sqla.exc.IntegrityError as integrity_error:
      raise datastore.AlreadyExistsError(
          'Suggest Op with name %s already exists.' %
          operation.name) from integrity_error

  def get_suggestion_operation(self,
                               operation_name: str) -> operations_pb2.Operation:
    query = sqla.select([self._suggestion_operations_table]).where(
        self._suggestion_operations_table.c.operation_name == operation_name)
    result = self._connection.execute(query)
    row = result.fetchone()

    if not row:
      raise datastore.NotFoundError('Failed to find suggest op name: %s' %
                                    operation_name)

    return operations_pb2.Operation.FromString(row['serialized_op'])

  def update_suggestion_operation(
      self, operation: operations_pb2.Operation
  ) -> resources.SuggestionOperationResource:
    resource = resources.SuggestionOperationResource.from_name(operation.name)

    exists_query = sqla.select([self._suggestion_operations_table])
    exists_query = exists_query.where(
        self._suggestion_operations_table.c.operation_name == operation.name)
    exists_query = sqla.exists(exists_query).select()
    exists = self._connection.execute(exists_query).fetchone()[0]
    if not exists:
      raise datastore.NotFoundError('Suggest op %s does not exist.' %
                                    operation.name)

    query = sqla.update(self._suggestion_operations_table).where(
        self._suggestion_operations_table.c.operation_name ==
        operation.name).values(
            operation_name=operation.name,
            owner_id=resource.owner_id,
            client_id=resource.client_id,
            operation_number=resource.operation_number,
            serialized_op=operation.SerializeToString())
    self._connection.execute(query)
    return resource

  def list_suggestion_operations(
      self,
      owner_name: str,
      client_id: str,
      filter_fn: Optional[Callable[[operations_pb2.Operation], bool]] = None
  ) -> List[operations_pb2.Operation]:
    owner_resource = resources.OwnerResource.from_name(owner_name)
    query = sqla.select([self._suggestion_operations_table])
    query = query.where(
        self._suggestion_operations_table.c.owner_id == owner_resource.owner_id)
    query = query.where(
        self._suggestion_operations_table.c.client_id == client_id)

    exists_query = sqla.exists(query).select()
    exists = self._connection.execute(exists_query).fetchone()[0]
    if not exists:
      raise datastore.NotFoundError('Could not find (owner_name, client_id):',
                                    (owner_name, client_id))

    result = self._connection.execute(query)
    all_ops = [
        operations_pb2.Operation.FromString(row['serialized_op'])
        for row in result
    ]

    if filter_fn is not None:
      output_list = []
      for op in all_ops:
        if filter_fn(op):
          output_list.append(op)
      return output_list
    else:
      return all_ops

  def max_suggestion_operation_number(self, owner_name: str,
                                      client_id: str) -> int:
    resource = resources.OwnerResource.from_name(owner_name)

    exists_query = sqla.select([self._suggestion_operations_table])
    exists_query = exists_query.where(
        self._suggestion_operations_table.c.owner_id == resource.owner_id)
    exists_query = exists_query.where(
        self._suggestion_operations_table.c.client_id == client_id)
    exists_query = sqla.exists(exists_query).select()
    exists = self._connection.execute(exists_query).fetchone()[0]
    if not exists:
      raise datastore.NotFoundError('Could not find (owner_name, client_id):',
                                    (owner_name, client_id))

    query = sqla.select([
        sqla.func.max(
            self._suggestion_operations_table.c.operation_number,
            type_=sqla.INT)
    ])
    query = query.where(
        self._suggestion_operations_table.c.owner_id == resource.owner_id)
    query = query.where(
        self._suggestion_operations_table.c.client_id == client_id)
    return self._connection.execute(query).fetchone()[0]

  def create_early_stopping_operation(
      self, operation: vizier_oss_pb2.EarlyStoppingOperation
  ) -> resources.EarlyStoppingOperationResource:
    resource = resources.EarlyStoppingOperationResource.from_name(
        operation.name)
    query = self._early_stopping_operations_table.insert().values(
        operation_name=operation.name,
        owner_id=resource.owner_id,
        study_id=resource.study_id,
        trial_id=resource.trial_id,
        serialized_op=operation.SerializeToString())
    try:
      self._connection.execute(query)
      return resource
    except sqla.exc.IntegrityError as integrity_error:
      raise datastore.AlreadyExistsError(
          'Early stopping op with name %s already exists.' %
          operation.name) from integrity_error

  def get_early_stopping_operation(
      self, operation_name: str) -> vizier_oss_pb2.EarlyStoppingOperation:
    query = sqla.select([self._early_stopping_operations_table])
    query = query.where(self._early_stopping_operations_table.c.operation_name
                        == operation_name)
    result = self._connection.execute(query)
    row = result.fetchone()

    if not row:
      raise datastore.NotFoundError(
          'Failed to find early stopping op name: %s' % operation_name)

    return vizier_oss_pb2.EarlyStoppingOperation.FromString(
        row['serialized_op'])

  def update_early_stopping_operation(
      self, operation: vizier_oss_pb2.EarlyStoppingOperation
  ) -> resources.EarlyStoppingOperationResource:
    resource = resources.EarlyStoppingOperationResource.from_name(
        operation.name)

    exists_query = sqla.select([self._early_stopping_operations_table])
    exists_query = exists_query.where(self._early_stopping_operations_table.c
                                      .operation_name == operation.name)
    exists_query = sqla.exists(exists_query).select()
    exists = self._connection.execute(exists_query).fetchone()[0]
    if not exists:
      raise datastore.NotFoundError('Early stopping op %s does not exist.' %
                                    operation.name)

    query = sqla.update(self._early_stopping_operations_table).where(
        self._early_stopping_operations_table.c.operation_name ==
        operation.name).values(
            operation_name=operation.name,
            owner_id=resource.owner_id,
            study_id=resource.study_id,
            trial_id=resource.trial_id,
            serialized_op=operation.SerializeToString())
    self._connection.execute(query)
    return resource

  def update_metadata(
      self,
      study_name: str,
      study_metadata: Iterable[key_value_pb2.KeyValue],
      trial_metadata: Iterable[datastore._KeyValuePlus],  # pylint:disable=protected-access
  ) -> None:
    """Store the supplied metadata into the SQL database."""
    # TODO: Add exception logic when this function is updated.
    s_resource = resources.StudyResource.from_name(study_name)
    logging.debug('database.update_metadata s_resource= %s', s_resource)

    # Obtain original study.
    get_study_query = sqla.select([
        self._studies_table
    ]).where(self._studies_table.c.study_name == study_name)
    study_result = self._connection.execute(get_study_query)
    row = study_result.fetchone()
    original_study = study_pb2.Study.FromString(row['serialized_study'])

    # Update the study with new study_metadata and update database.
    original_study.study_spec.ClearField('metadata')
    for metadata in study_metadata:
      original_study.study_spec.metadata.append(metadata)
    update_study_query = sqla.update(self._studies_table).where(
        self._studies_table.c.study_name == study_name).values(
            serialized_study=original_study.SerializeToString())
    self._connection.execute(update_study_query)

    # Store trial-related metadata in the database. We first create a dict of
    # the relevant `trial_resources` that will be touched.   We clear them, then
    # loop through the metadata, converting to protos.
    trial_resources: Dict[str, resources.TrialResource] = {}
    for metadata in trial_metadata:
      clear_metadata_bool = False
      if metadata.trial_id in trial_resources:
        t_resource = trial_resources[metadata.trial_id]
      else:
        # If we don't have a t_resource entry already, create one and clear the
        # relevant Trial's metadata.
        t_resource = s_resource.trial_resource(metadata.trial_id)
        trial_resources[metadata.trial_id] = t_resource
        clear_metadata_bool = True

      # Obtain original trial.
      trial_name = t_resource.name
      original_trial_query = sqla.select([
          self._trials_table
      ]).where(self._trials_table.c.trial_name == trial_name)
      trial_result = self._connection.execute(original_trial_query)
      row = trial_result.fetchone()
      original_trial = study_pb2.Trial.FromString(row['serialized_trial'])

      # Edit trial metadata and update database.
      if clear_metadata_bool:
        original_trial.ClearField('metadata')
      original_trial.metadata.append(metadata.k_v)
      update_trial_query = sqla.update(self._trials_table).where(
          self._trials_table.c.trial_name == trial_name).values(
              serialized_trial=original_trial.SerializeToString())
      self._connection.execute(update_trial_query)



"""Algorithm core modules."""

from vizier._src.algorithms.core.abstractions import CompletedTrials
from vizier._src.algorithms.core.abstractions import Designer
from vizier._src.algorithms.core.abstractions import PartiallySerializableDesigner
from vizier._src.algorithms.core.abstractions import SerializableDesigner
from vizier._src.algorithms.core.optimizers import BranchSelection
from vizier._src.algorithms.core.optimizers import BranchSelector
from vizier._src.algorithms.core.optimizers import BranchThenMaximizer
from vizier._src.algorithms.core.optimizers import GradientFreeMaximizer



"""Setup for pip package."""

from setuptools import find_namespace_packages
from setuptools import setup


def _get_version():
  with open('vizier/__init__.py') as fp:
    for line in fp:
      if line.startswith('__version__'):
        g = {}
        exec(line, g)  # pylint: disable=exec-used
        return g['__version__']
    raise ValueError('`__version__` not defined in `vizier/__init__.py`')


def _parse_requirements(requirements_txt_path):
  with open(requirements_txt_path) as fp:
    return fp.read().splitlines()


_VERSION = _get_version()

setup(
    name='google-vizier',
    version=_VERSION,
    url='https://github.com/google/vizier',
    license='Apache License 2.0',
    author='Vizier Team',
    description='Vizier: Distributed service framework for blackbox optimization and research.',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    author_email='oss-vizier-dev@google.com',
    # Contained modules and scripts.
    packages=find_namespace_packages(
        include=['vizier*'], exclude=['*_test.py', 'examples']),
    install_requires=_parse_requirements('requirements.txt'),
    extras_require={
        'jax': _parse_requirements('requirements-jax.txt'),
        'tf': _parse_requirements('requirements-tf.txt'),
        'misc': _parse_requirements('requirements-misc.txt')
    },
    # tests_require=_parse_requirements('requirements-test.txt'),
    requires_python='>=3.9',
    include_package_data=True,
    zip_safe=False,
    # PyPI package information.
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Intended Audience :: Education',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.9',
        'Topic :: Scientific/Engineering :: Mathematics',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Software Development :: Libraries',
    ],
    keywords='ai machine learning hyperparameter blackbox optimization framework',
)



"""Benchmark classes for Vizier."""

from vizier._src.benchmarks.experimenters.experimenter import Experimenter
from vizier._src.benchmarks.experimenters.numpy_experimenter import NumpyExperimenter
from vizier._src.benchmarks.experimenters.synthetic import bbob



"""Compatibility method comparison for protobufs.

Note that using this method can be unstable and hard to debug.
Provided here for compatibility.
"""

# TODO: Add pytypes.
# TODO: Fix this library.


# pylint: disable=invalid-name
def assertProto2Equal(self, x, y):
  self.assertEqual(x.SerializeToString(), y.SerializeToString())


def assertProto2Contains(self, x, y):
  del self
  if not isinstance(x, str):
    x = str(x)
  if x not in str(y):
    pass


def assertProto2SameElements(*args, **kwargs):
  del args, kwargs
  pass



"""Tests for random."""

from vizier import pyvizier as vz
from vizier._src.algorithms.designers import random
from vizier._src.algorithms.testing import test_runners
from vizier.testing import test_studies

from absl.testing import absltest


class RandomTest(absltest.TestCase):

  def test_on_flat_space(self):
    config = vz.StudyConfig(test_studies.flat_space_with_all_types())
    designer = random.RandomDesigner(config.search_space, seed=None)
    self.assertLen(
        test_runners.run_with_random_metrics(
            designer, config, iters=50, batch_size=1), 50)

  def test_reproducible_random(self):
    config = vz.StudyConfig(test_studies.flat_space_with_all_types())
    designer = random.RandomDesigner(config.search_space, seed=5)
    t1 = designer.suggest(10)

    designer = random.RandomDesigner(config.search_space, seed=5)
    t2 = designer.suggest(10)
    self.assertEqual(t1, t2)


if __name__ == '__main__':
  absltest.main()



"""Random designer."""

from typing import Any, Optional, Sequence

import numpy as np
from vizier import algorithms as vza
from vizier import pyvizier as vz
from vizier.pyvizier import converters


class RandomDesigner(vza.Designer):
  """Random designer.

  This designer uniformly samples from the scaled search space.
  """

  def __init__(self,
               search_space: vz.SearchSpace,
               *,
               dtype: Any = np.float64,
               seed: Any):
    """Init.

    Args:
      search_space: Must be a flat search space.
      dtype:
      seed: Any valid seed for np.random.RandomState.
    """
    if search_space.is_conditional:
      raise ValueError(
          f'This designer {self} does not support conditional search.')

    def create_input_converter(pc):
      return converters.DefaultModelInputConverter(
          pc, scale=True, max_discrete_indices=0, float_dtype=dtype)

    self._converter = converters.DefaultTrialConverter(
        [create_input_converter(pc) for pc in search_space.parameters])

    self._rng = np.random.RandomState(seed)

  def update(self, _) -> None:
    pass

  def suggest(self,
              count: Optional[int] = None) -> Sequence[vz.TrialSuggestion]:
    """Make new suggestions.

    Args:
      count: Makes best effort to generate this many suggestions. If None,
        suggests as many as the algorithm wants.

    Returns:
      New suggestions.
    """
    count = count or 1
    sample = dict()
    for name, spec in self._converter.output_specs.items():
      if spec.type == converters.NumpyArraySpecType.DISCRETE:
        sample[name] = self._rng.random_integers(
            spec.bounds[0], spec.bounds[1] - spec.num_oovs, size=[count, 1])
      elif spec.type == converters.NumpyArraySpecType.CONTINUOUS:
        # Trial-Numpy converter was configured to scale values to [0, 1].
        # Sample uniformly in that range and convert back to Trials.
        sample[name] = self._rng.random([count, 1])
      else:
        raise ValueError(f'Unsupported type: {spec.type} in {spec}')
    return [
        vz.TrialSuggestion(p) for p in self._converter.to_parameters(sample)
    ]



"""Test study generator."""

from typing import Collection
from vizier import pyvizier as vz


def flat_space_with_all_types() -> vz.SearchSpace:
  """Search space with all parameter types."""

  space = vz.SearchSpace()
  root = space.select_root()
  root.add_float_param('lineardouble', -1., 2.)
  root.add_float_param('logdouble', 1e-4, 1e2, scale_type=vz.ScaleType.LOG)
  root.add_int_param('integer', -2, 2)
  root.add_categorical_param('categorical', ['a', 'aa', 'aaa'])
  root.add_bool_param('boolean')
  root.add_discrete_param('discrete_double', [-.5, 1.0, 1.2])
  root.add_discrete_param(
      'discrete_logdouble', [1e-5, 1e-2, 1e-1], scale_type=vz.ScaleType.LOG)
  root.add_discrete_param('discrete_int', [-1, 1, 2])

  return space


def metrics_objective_goals() -> Collection[vz.MetricInformation]:
  return [
      vz.MetricInformation('gain', goal=vz.ObjectiveMetricGoal.MAXIMIZE),
      vz.MetricInformation('loss', goal=vz.ObjectiveMetricGoal.MINIMIZE),
      vz.MetricInformation(
          'auc',
          goal=vz.ObjectiveMetricGoal.MAXIMIZE,
          min_value=0.0,
          max_value=1.0),
      vz.MetricInformation(
          'crossentropy', goal=vz.ObjectiveMetricGoal.MINIMIZE, min_value=0.0),
  ]


def metrics_all_unconstrained() -> Collection[vz.MetricInformation]:
  return [
      vz.MetricInformation('gain', goal=vz.ObjectiveMetricGoal.MAXIMIZE),
      vz.MetricInformation('loss', goal=vz.ObjectiveMetricGoal.MINIMIZE),
      vz.MetricInformation(
          'gt2', goal=vz.ObjectiveMetricGoal.MAXIMIZE, safety_threshold=2.0),
      vz.MetricInformation(
          'lt2', goal=vz.ObjectiveMetricGoal.MINIMIZE, safety_threshold=2.0),
  ]


def metrics_all_constrained() -> Collection[vz.MetricInformation]:
  return [
      vz.MetricInformation(
          'auc',
          goal=vz.ObjectiveMetricGoal.MAXIMIZE,
          min_value=0.0,
          max_value=1.0),
      vz.MetricInformation(
          'crossentropy', goal=vz.ObjectiveMetricGoal.MINIMIZE, min_value=0.0),
      vz.MetricInformation(
          'gt2',
          goal=vz.ObjectiveMetricGoal.MAXIMIZE,
          safety_threshold=2.0,
          min_value=-1.0,
          max_value=5.0),
      vz.MetricInformation(
          'lt2',
          goal=vz.ObjectiveMetricGoal.MINIMIZE,
          safety_threshold=2.0,
          min_value=-1.0,
          max_value=5.0),
  ]



"""ParameterConfig wraps ParameterConfig and ParameterSpec protos."""

import collections
import copy
import enum
import math
from typing import Generator, List, Optional, Sequence, Tuple, Union

from absl import logging
import attr

from vizier._src.pyvizier.shared import trial

ExternalType = trial.ExternalType


class ParameterType(enum.Enum):
  """Valid Values for ParameterConfig.type."""
  DOUBLE = 'DOUBLE'
  INTEGER = 'INTEGER'
  CATEGORICAL = 'CATEGORICAL'
  DISCRETE = 'DISCRETE'

  def is_numeric(self) -> bool:
    return self in [self.DOUBLE, self.INTEGER, self.DISCRETE]


class ScaleType(enum.Enum):
  """Valid Values for ParameterConfig.scale_type."""
  LINEAR = 'LINEAR'
  LOG = 'LOG'
  REVERSE_LOG = 'REVERSE_LOG'
  UNIFORM_DISCRETE = 'UNIFORM_DISCRETE'


# A sequence of possible internal parameter values.
MonotypeParameterSequence = Union[Sequence[Union[int, float]], Sequence[str]]
MonotypeParameterList = Union[List[Union[int, float]], List[str]]


def _validate_bounds(bounds: Union[Tuple[int, int], Tuple[float, float]]):
  """Validates the bounds."""
  if len(bounds) != 2:
    raise ValueError('Bounds must have length 2. Given: {}'.format(bounds))
  lower = bounds[0]
  upper = bounds[1]
  if not all([math.isfinite(v) for v in (lower, upper)]):
    raise ValueError(
        'Both "lower" and "upper" must be finite. Given: (%f, %f)' %
        (lower, upper))
  if lower > upper:
    raise ValueError(
        'Lower cannot be greater than upper: given lower={} upper={}'.format(
            lower, upper))


def _get_feasible_points_and_bounds(
    feasible_values: Sequence[float]
) -> Tuple[List[float], Union[Tuple[int, int], Tuple[float, float]]]:
  """Validates and converts feasible values to floats."""
  if not all([math.isfinite(p) for p in feasible_values]):
    raise ValueError('Feasible values must all be finite. Given: {}' %
                     feasible_values)

  feasible_points = list(sorted(feasible_values))
  bounds = (feasible_points[0], feasible_points[-1])
  return feasible_points, bounds


def _get_categories(categories: Sequence[str]) -> List[str]:
  """Returns the categories."""
  return sorted(list(categories))


def _get_default_value(
    param_type: ParameterType,
    default_value: Union[float, int, str]) -> Union[float, int, str]:
  """Validates and converts the default_value to the right type."""
  if (param_type in (ParameterType.DOUBLE, ParameterType.DISCRETE) and
      (isinstance(default_value, float) or isinstance(default_value, int))):
    return float(default_value)
  elif (param_type == ParameterType.INTEGER and
        (isinstance(default_value, float) or isinstance(default_value, int))):
    if isinstance(default_value, int):
      return default_value
    else:
      # Check if the float rounds nicely.
      default_int_value = round(default_value)
      if not math.isclose(default_value, default_int_value):
        raise ValueError('default_value for an INTEGER parameter should be an '
                         'integer, got float: [{}]'.format(default_value))
      return default_int_value
  elif (param_type == ParameterType.CATEGORICAL and
        isinstance(default_value, str)):
    return default_value
  raise ValueError(
      'default_value has an incorrect type. ParameterType has type {}, '
      'but default_value has type {}'.format(param_type.name,
                                             type(default_value)))


@attr.s(auto_attribs=True, frozen=True, init=True, slots=True)
class ParameterConfig:
  """A Vizier ParameterConfig.

  Use ParameterConfig.factory to create a valid instance.
  """
  _name: str = attr.ib(
      init=True, validator=attr.validators.instance_of(str), kw_only=True)
  _type: ParameterType = attr.ib(
      init=True,
      validator=attr.validators.instance_of(ParameterType),
      repr=lambda v: v.name if v is not None else 'None',
      kw_only=True)
  # Only one of _feasible_values, _bounds will be set at any given time.
  _bounds: Optional[Union[Tuple[int, int], Tuple[float, float]]] = attr.ib(
      init=True,
      validator=attr.validators.optional(
          attr.validators.deep_iterable(
              member_validator=attr.validators.instance_of((int, float)),
              iterable_validator=attr.validators.instance_of(tuple))),
      kw_only=True)
  _feasible_values: Optional[MonotypeParameterList] = attr.ib(
      init=True,
      validator=attr.validators.optional(
          attr.validators.deep_iterable(
              member_validator=attr.validators.instance_of((int, float, str)),
              iterable_validator=attr.validators.instance_of((list, tuple)))),
      kw_only=True)
  _scale_type: Optional[ScaleType] = attr.ib(
      init=True,
      validator=attr.validators.optional(
          attr.validators.instance_of(ScaleType)),
      repr=lambda v: v.name if v is not None else 'None',
      kw_only=True)
  _default_value: Optional[Union[float, int, str]] = attr.ib(
      init=True,
      validator=attr.validators.optional(
          attr.validators.instance_of((float, int, str))),
      kw_only=True)
  _external_type: ExternalType = attr.ib(
      init=True,
      converter=lambda v: v or ExternalType.INTERNAL,
      validator=attr.validators.optional(
          attr.validators.instance_of(ExternalType)),
      repr=lambda v: v.name if v is not None else 'None',
      kw_only=True)
  # Parent values for this ParameterConfig. If set, then this is a child
  # ParameterConfig.
  _matching_parent_values: Optional[MonotypeParameterList] = attr.ib(
      init=True,
      validator=attr.validators.optional(
          attr.validators.deep_iterable(
              member_validator=attr.validators.instance_of((int, float, str)),
              iterable_validator=attr.validators.instance_of((list, tuple)))),
      kw_only=True)
  # Children ParameterConfig. If set, then this is a parent ParameterConfig.
  _child_parameter_configs: Optional[List['ParameterConfig']] = attr.ib(
      init=True, kw_only=True)

  # Pytype treats instances of EnumTypeWrapper as types, but they can't be
  # evaluated at runtime, so a Union[] of proto enums has to be a forward
  # reference below.
  @classmethod
  def factory(
      cls,
      name: str,
      *,
      bounds: Optional[Union[Tuple[int, int], Tuple[float, float]]] = None,
      feasible_values: Optional[MonotypeParameterSequence] = None,
      children: Optional[Sequence[Tuple[MonotypeParameterSequence,
                                        'ParameterConfig']]] = None,
      scale_type: Optional[ScaleType] = None,
      default_value: Optional[Union[float, int, str]] = None,
      external_type: Optional[ExternalType] = ExternalType.INTERNAL
  ) -> 'ParameterConfig':
    """Factory method.

    Args:
      name: The parameter's name. Cannot be empty.
      bounds: REQUIRED for INTEGER or DOUBLE type. Specifies (min, max). The
        type of (min, max) determines the created ParameterConfig's type.
      feasible_values: REQUIRED for DISCRETE or CATEGORICAL type. The elements'
        type determines the created ParameterConfig's type.
      children: sequence of tuples formatted as: (matching_parent_values,
        ParameterConfig). See
        cs/learning_vizier.service.ParameterConfig.child_parameter_configs for
        details. ONLY THE TYPES ARE VALIDATED. If the child ParameterConfig
        protos already have parent values set, they will be overridden by the
        provided matching_parent_values.
      scale_type: Scaling to be applied. NOT VALIDATED.
      default_value: A default value for the Parameter.
      external_type: An annotation indicating the type this parameter should be
        cast to.

    Returns:
      A ParameterConfig object which wraps a partially validated proto.

    Raises:
      ValueError: Exactly one of feasible_values and bounds must be convertible
        to Boolean true. Bounds and numeric feasible_values must be finite.
        Bounds and feasible_values, if provided, must consist of
        elements of the same type.
      TypeError: If children's matching_parent_values are not compatible with
        the ParameterConfig being created.
    """
    if not name:
      raise ValueError('Parameter name cannot be empty.')

    if bool(feasible_values) == bool(bounds):
      raise ValueError(
          'While creating Parameter with name={}: exactly one of '
          '"feasible_values" or "bounds" must be provided, but given '
          'feasible_values={} and bounds={}.'.format(name, feasible_values,
                                                     bounds))
    if feasible_values:
      if len(set(feasible_values)) != len(feasible_values):
        counter = collections.Counter(feasible_values)
        duplicate_dict = {k: v for k, v in counter.items() if v > 1}
        raise ValueError(
            'Feasible values cannot have duplicates: {}'.format(duplicate_dict))
      if all(isinstance(v, (float, int)) for v in feasible_values):
        inferred_type = ParameterType.DISCRETE
        feasible_values, bounds = _get_feasible_points_and_bounds(
            feasible_values)
      elif all(isinstance(v, str) for v in feasible_values):
        inferred_type = ParameterType.CATEGORICAL
        feasible_values = _get_categories(feasible_values)
      else:
        raise ValueError(
            'Feasible values must all be numeric or strings. Given {}'.format(
                feasible_values))
    else:  # bounds were specified.
      if isinstance(bounds[0], int) and isinstance(bounds[1], int):
        inferred_type = ParameterType.INTEGER
        _validate_bounds(bounds)
      elif isinstance(bounds[0], float) and isinstance(bounds[1], float):
        inferred_type = ParameterType.DOUBLE
        _validate_bounds(bounds)
      else:
        raise ValueError(
            'Bounds must both be integers or doubles. Given: {}'.format(bounds))

    if default_value is not None:
      default_value = _get_default_value(inferred_type, default_value)

    pc = cls(
        name=name,
        type=inferred_type,
        bounds=bounds,
        feasible_values=feasible_values,
        scale_type=scale_type,
        default_value=default_value,
        external_type=external_type,
        matching_parent_values=None,
        child_parameter_configs=None)
    if children:
      pc = pc.add_children(children)
    return pc

  @property
  def name(self) -> str:
    return self._name

  @property
  def type(self) -> ParameterType:
    return self._type

  @property
  def external_type(self) -> ExternalType:
    return self._external_type

  @property
  def scale_type(self) -> Optional[ScaleType]:
    return self._scale_type

  @property
  def bounds(self) -> Union[Tuple[float, float], Tuple[int, int]]:
    """Returns the bounds, if set, or raises a ValueError."""
    if self.type == ParameterType.CATEGORICAL:
      raise ValueError('Accessing bounds of a categorical parameter: %s' %
                       self.name)
    return self._bounds

  @property
  def matching_parent_values(self) -> MonotypeParameterList:
    """Returns the matching parent values, if this is a child parameter."""
    if not self._matching_parent_values:
      return []
    return copy.copy(self._matching_parent_values)

  @property
  def child_parameter_configs(self) -> List['ParameterConfig']:
    if not self._child_parameter_configs:
      return []
    return copy.deepcopy(self._child_parameter_configs)

  def _del_child_parameter_configs(self):
    """Deletes the current child ParameterConfigs."""
    object.__setattr__(self, '_child_parameter_configs', None)

  @property
  def clone_without_children(self) -> 'ParameterConfig':
    """Returns the clone of self, without child_parameter_configs."""
    clone = copy.deepcopy(self)
    clone._del_child_parameter_configs()  # pylint: disable='protected-access'
    return clone

  @property
  def feasible_values(self) -> Union[List[int], List[float], List[str]]:
    if self.type in (ParameterType.DISCRETE, ParameterType.CATEGORICAL):
      if not self._feasible_values:
        return []
      return copy.copy(self._feasible_values)
    elif self.type == ParameterType.INTEGER:
      return list(range(self.bounds[0], self.bounds[1] + 1))
    raise ValueError('feasible_values is invalid for type: %s' % self.type)

  @property
  def default_value(self) -> Optional[Union[int, float, str]]:
    """Returns the default value, or None if not set."""
    return self._default_value

  def _set_matching_parent_values(self,
                                  parent_values: MonotypeParameterSequence):
    """Sets the given matching parent values in this object, without validation.

    Args:
      parent_values: Parent values for which this child ParameterConfig is
        active. Existing values will be replaced.
    """
    object.__setattr__(self, '_matching_parent_values', list(parent_values))

  def _set_child_parameter_configs(self, children: List['ParameterConfig']):
    """Sets the given child ParameterConfigs in this object, without validation.

    Args:
      children: The children to set in this object. Existing children will be
        replaced.
    """
    object.__setattr__(self, '_child_parameter_configs', children)

  def add_children(
      self, new_children: Sequence[Tuple[MonotypeParameterSequence,
                                         'ParameterConfig']]
  ) -> 'ParameterConfig':
    """Clones the ParameterConfig and adds new children to it.

    Args:
      new_children: A sequence of tuples formatted as: (matching_parent_values,
        ParameterConfig). If the child ParameterConfig have pre-existing parent
        values, they will be overridden.

    Returns:
      A parent parameter config, with children set.

    Raises:
      ValueError: If the child configs are invalid
      TypeError: If matching parent values are invalid
    """
    parent = copy.deepcopy(self)
    if not new_children:
      return parent

    for child_pair in new_children:
      if len(child_pair) != 2:
        raise ValueError('Each element in new_children must be a tuple of '
                         '(Sequence of valid parent values,  ParameterConfig),'
                         ' given: {}'.format(child_pair))

    logging.debug('add_children: new_children=%s', new_children)
    child_parameter_configs = parent.child_parameter_configs
    for unsorted_parent_values, child in new_children:
      parent_values = sorted(unsorted_parent_values)
      child_copy = copy.deepcopy(child)
      if parent.type == ParameterType.DISCRETE:
        if not all(isinstance(v, (float, int)) for v in parent_values):
          raise TypeError('Parent is DISCRETE-typed, but a child is specifying '
                          'one or more non float/int parent values: child={} '
                          ', parent_values={}'.format(child, parent_values))
        child_copy._set_matching_parent_values(parent_values)  # pylint: disable='protected-access'
      elif parent.type == ParameterType.CATEGORICAL:
        if not all(isinstance(v, str) for v in parent_values):
          raise TypeError('Parent is CATEGORICAL-typed, but a child is '
                          'specifying one or more non float/int parent values: '
                          'child={}, parent_values={}'.format(
                              child, parent_values))
        child_copy._set_matching_parent_values(parent_values)  # pylint: disable='protected-access'
      elif parent.type == ParameterType.INTEGER:
        # Allow {int, float}->float conversion but block str->float conversion.
        int_values = [int(v) for v in parent_values]
        if int_values != parent_values:
          raise TypeError(
              'Parent is INTEGER-typed, but a child is specifying one or more '
              'non-integral parent values: {}'.format(parent_values))
        child_copy._set_matching_parent_values(int_values)  # pylint: disable='protected-access'
      else:
        raise ValueError('DOUBLE type cannot have child parameters')
      child_parameter_configs.extend([child_copy])
    parent._set_child_parameter_configs(child_parameter_configs)  # pylint: disable='protected-access'
    return parent

  def continuify(self) -> 'ParameterConfig':
    """Returns a newly created DOUBLE parameter with the same range."""
    if self.type == ParameterType.DOUBLE:
      return copy.deepcopy(self)
    elif not ParameterType.is_numeric(self.type):
      raise ValueError(
          'Cannot convert a non-numeric parameter to DOUBLE: {}'.format(self))
    elif self._child_parameter_configs:
      raise ValueError(
          'Cannot convert a parent parameter to DOUBLE: {}'.format(self))

    scale_type = self.scale_type
    if scale_type == ScaleType.UNIFORM_DISCRETE:
      logging.log_every_n(
          logging.WARNING,
          'Converting a UNIFORM_DISCRETE scaled discrete parameter '
          'to DOUBLE: %s', 10, self)
      scale_type = None

    default_value = self.default_value
    if default_value is not None:
      default_value = float(default_value)
    return ParameterConfig.factory(
        self.name,
        bounds=(float(self.bounds[0]), float(self.bounds[1])),
        scale_type=scale_type,
        default_value=default_value)

  @classmethod
  def merge(cls, one: 'ParameterConfig',
            other: 'ParameterConfig') -> 'ParameterConfig':
    """Merge two ParameterConfigs.

    Args:
      one: ParameterConfig with no child parameters.
      other: Must have the same type as one, and may not have child parameters.

    Returns:
      For Categorical, Discrete or Integer ParameterConfigs, the resulting
      config will be the union of all feasible values.
      For Double ParameterConfigs, the resulting config will have [min_value,
      max_value] set to the smallest and largest bounds.

    Raises:
      ValueError: If any of the input configs has child parameters, or if
        the two parameters have different types.
    """
    if one.child_parameter_configs or other.child_parameter_configs:
      raise ValueError(
          'Cannot merge parameters with child_parameter_configs: %s and %s' %
          one, other)
    if one.type != other.type:
      raise ValueError('Type conflicts between {} and {}'.format(
          one.type.name, other.type.name))
    if one.scale_type != other.scale_type:
      logging.warning('Scale type conflicts while merging %s and %s', one,
                      other)

    if one.type in (ParameterType.CATEGORICAL, ParameterType.DISCRETE):
      new_feasible_values = list(
          set(one.feasible_values + other.feasible_values))
      return ParameterConfig.factory(
          name=one.name,
          feasible_values=new_feasible_values,
          scale_type=one.scale_type)
    elif one.type in (ParameterType.INTEGER, ParameterType.DOUBLE):
      original_min, original_max = one.bounds
      other_min, other_max = other.bounds
      new_bounds = (min(original_min, other_min), max(original_max, other_max))
      return ParameterConfig.factory(
          name=one.name, bounds=new_bounds, scale_type=one.scale_type)
    raise ValueError('Unknown type {}. This is currently'
                     'an unreachable code.'.format(one.type))

  def traverse(
      self,
      show_children: bool = False) -> Generator['ParameterConfig', None, None]:
    """DFS Generator for parameter configs.

    Args:
      show_children: If True, every generated ParameterConfig has
        child_parameter_configs. For example, if 'foo' has two child configs
        'bar1' and 'bar2', then traversing 'foo' with show_children=True
        generates (foo, with bar1,bar2 as children), (bar1), and (bar2). If
        show_children=False, it generates (foo, without children), (bar1), and
        (bar2).

    Yields:
      DFS on all parameter configs.
    """
    if show_children:
      yield self
    else:
      yield self.clone_without_children
    for child in self.child_parameter_configs:
      yield from child.traverse(show_children)

  def contains(
      self, value: Union[trial.ParameterValueTypes,
                         trial.ParameterValue]) -> bool:
    """Check if the `value` is a valid value for this parameter config."""
    if not isinstance(value, trial.ParameterValue):
      value = trial.ParameterValue(value)

    if self.type == ParameterType.DOUBLE:
      return self.bounds[0] <= value.as_float and value.as_float <= self.bounds[
          1]
    elif self.type == ParameterType.INTEGER:
      if value.as_int != value.as_float:
        return False
      return self.bounds[0] <= value.as_int and value.as_int <= self.bounds[1]
    elif self.type == ParameterType.DISCRETE:
      return value.as_float in self.feasible_values
    elif self.type == ParameterType.CATEGORICAL:
      return value.as_str in self.feasible_values
    else:
      raise NotImplementedError(f'Cannot determine whether {value} is feasible'
                                f'for Unknown parameter type {self.type}.\n'
                                f'Full config: {repr(self)}')

  @property
  def num_feasible_values(self) -> Union[float, int]:
    if self.type == ParameterType.DOUBLE:
      return float('inf')
    elif self.type == ParameterType.INTEGER:
      return self.bounds[1] - self.bounds[0] + 1
    else:
      return len(self.feasible_values)
